/*
 *  cir.c : A SIS CIR driver for Linux.
 *  This software may be used and distributed according to the terms
 *  of the GNU Public License (GPL), incorporated herein by reference.
 *  Drivers based on this skeleton fall under the GPL and must retain
 *  the authorship (implicit copyright) notice.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/pci.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/irq.h>

#include "cir.h"

static struct cir_status cir;
static int cir_irq = CIR_IRQ;

#define DEVICE_NAME "cir"
#ifdef  MODULE
MODULE_PARM(cir_irq, "i");
#endif

static unsigned int cir_poll(struct file *file, poll_table * wait);
static int cir_fasync(int fd, struct file *filp, int on);
static int cir_close(struct inode *inode, struct file *file);
static int cir_open(struct inode *inode, struct file *file);
static ssize_t cir_write(struct file *file,
                         const char *buffer, size_t count, loff_t * ppos);
static ssize_t cir_read(struct file *file,
                        char *buffer, size_t count, loff_t * ppos);
static int              data_no;
//#define cir_inc(x) (x < MAX_NUM-1) ? x+1 : 0

void init_silitek();
void init_btc();
void init_forward();

struct file_operations cir_fops = {
        read:           cir_read,
        write:          cir_write,
        poll:           cir_poll,
        open:           cir_open,
        release:        cir_close,
        fasync:         cir_fasync,
};

static struct miscdevice bus_cir = {
        CIR_MINOR, "cir", &cir_fops
};

static void cir_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	u8 io_value,Data1,Data2,tmp[3];
	int i, len = 0;
	
	len = inb(CIR_RSR_PORT);

//	MSG("data length reveived = %d bytes\n",len);
//printk("data length reveived = %d bytes\n",len);
//        printk("cir.tail %x cir.head %x\n",cir.tail,cir.head);
//        len=inb(CIR_RSR_PORT);
	while(len>0){
	        tmp[0]=inb(CIR_DATA_PORT);
	        tmp[1]=inb(CIR_DATA_PORT);
//	        printk("rece data %x   %x \n",tmp[0],tmp[1]);
	        
	        if(tmp[0]==0x3f){
	           tmp[2]=inb(CIR_DATA_PORT);
	           for(i=0;i<3;i++)
	           {
	              cir.buff[cir.tail]=tmp[i];
	              cir.tail = (cir.tail + 1) % CIR_FIFO_SIZE;
	           }   
            	   len = len - 3; 
//            	    printk("lenfor3 --- %d \n",len);
            	    
            	   data_no=data_no+3;    
	        }
                else 
                {
                   for(i=0;i<2;i++)
                   {
                     cir.buff[cir.tail]=tmp[i];
                     cir.tail = ( cir.tail + 1 ) % CIR_FIFO_SIZE;
                   }
                   len = len - 2;
//                   printk("lenfor2 --- %d \n",len);
                   cir.buff[cir.tail]=0;
                   cir.tail = ( cir.tail + 1 ) % CIR_FIFO_SIZE;
                   data_no=data_no+3;
                }
                if(cir.tail==cir.head)   // FIFO empty ?
                    cir.head=( cir.head + 3 ) % CIR_FIFO_SIZE;
//printk("len %d \n",len);
	}    //end of while

io_value = inb(CIR_ISR_PORT);
//printk("clear status%x \n",io_value);
	/* clear CIR status port */
//       while ((io_value = inb(CIR_ISR_PORT))!= 0x0){
         do{
//	        printk("clear status 1--- port312 %x \n",io_value);	
		if(io_value & 0x40){
		   
//		   printk(" overthreshold \n");
		   
		   len=inb(CIR_RSR_PORT);
       	           
       	           while(len>0){
	                tmp[0]=inb(CIR_DATA_PORT);
	                tmp[1]=inb(CIR_DATA_PORT);
	                if(tmp[0]==0x3f){
	                    tmp[2]=inb(CIR_DATA_PORT);
	                    for(i=0;i<3;i++)
	                    {
	                         cir.buff[cir.tail]=tmp[i];
	                         cir.tail = (cir.tail + 1) % CIR_FIFO_SIZE;
	                    }   
            	            len -= 3;     
	                } 
                        else 
                        {
                           for(i=0;i<2;i++)
                           {
                              cir.buff[cir.tail]=tmp[i];
                              cir.tail = ( cir.tail + 1 ) % CIR_FIFO_SIZE;
                           }
                          len -= 2;
                          cir.buff[cir.tail]=0;
                          cir.tail = ( cir.tail + 1 ) % CIR_FIFO_SIZE;
                        } //end for if
                       if(cir.tail==cir.head)   // FIFO empty ?
                          cir.head=( cir.head + 3 ) % CIR_FIFO_SIZE;
		   } //end of while 
                } //end of if
		io_value = inb(CIR_ISR_PORT); 
		outb(io_value, CIR_ISR_PORT);
		io_value = inb(CIR_ISR_PORT);
//          printk("clear status 2--- port312 %x \n",io_value);		
	} while(io_value!=0x0);
//printk("ret \n");
		/* wake up blocked cir_read command */
		wake_up_interruptible(&cir.wait);

	if (cir.fasyncptr)
		kill_fasync(&cir.fasyncptr, SIGIO, POLL_IN);
		
/*
 do {
     io_value = inb(CIR_ISR_PORT);
     if(io_value & 0x40){
         len=inb(0x315);
         printk("length %x \n",len);
 
         while(len--)  inb(0x310);
    }
    io_value = inb(CIR_ISR_PORT);
    outb(io_value, CIR_ISR_PORT);
    io_value = inb(CIR_ISR_PORT);
    printk("clear status 3 --- port312 %x \n",inb(0x312));
} while (io_value != 0x0);
*/
		
}

static int cir_fasync(int fd, struct file *filp, int on)
{
	int retval;
	retval = fasync_helper(fd, filp, on, &cir.fasyncptr);
	if (retval < 0)
		return retval;
	return 0;
}

/*
 * close access to the cir 
 */

static int cir_close(struct inode *inode, struct file *file)
{
	MSG("close siscir\n");
	cir_fasync(-1, file, 0);
	if (--cir.active)
		return 0;
	MOD_DEC_USE_COUNT;
	return 0;
}


/*
 * open access to the cir 
 */

static int cir_open(struct inode *inode, struct file *file)
{
	if (!cir.present)
		return -EINVAL;
	if (cir.active == 1)
		return -EAGAIN;
	if (cir.active++)
		return 0;

	MSG("open siscir\n");
	MOD_INC_USE_COUNT;
	MSG("open SiS CIR with IRQ %d \n", cir_irq);

	return 0;
}

/* 
 * writes are disallowed
 */

static ssize_t cir_write(struct file *file,
			 const char *buffer, size_t count, loff_t * ppos)
{
	return -EINVAL;
}

/*
 * read cir data.
 */

static ssize_t cir_read(struct file *file,
			char *buffer, size_t count, loff_t * ppos)
{
	int r,i;
	/* long flags; */

	if (count < 1) {
		return -EINVAL;
	}
//        printk("cir_read len %x \n",count);
	if ((r = verify_area(VERIFY_WRITE, buffer, count)))
		return -EINVAL;

	if (data_no==CIR_FIFO_SIZE){ // FIFO fulled ?
                /* nonblock read */
                if(file->f_flags & O_NONBLOCK)
	                return -EAGAIN;
		else{
               	/* block read */
			MSG("putting process with pid %u to sleep\n", current->pid);
                        interruptible_sleep_on(&cir.wait);
                        if (signal_pending(current)){
				MSG("pid %u got signal\n", current->pid);
                                return -ERESTARTSYS;
			}
		}
	}
	/* put user data */
	for(i=0; i < count ; i++)
		if ( data_no >0 ){ 	
			put_user(cir.buff[cir.head], buffer);
			cir.head = (cir.head + 1)%CIR_FIFO_SIZE;
			buffer++;
		}
		else  /* buffer empty */
			return i;
	if (data_no>=3) data_no=data_no	-3;
	return count ;		
}

/* 
 * poll for cir input 
 */

static unsigned int cir_poll(struct file *file, poll_table * wait)
{
	poll_wait(file, &cir.wait, wait);
//	if (cir.head != cir.tail)
        if(data_no <=CIR_FIFO_SIZE)
		return POLLIN | POLLRDNORM;
	return 0;
}

/*
 * Initiation of silitek cir keybaord 
 */
void init_silitek()
{
	outb(0x80, 0x313);
	outb(0x30, 0x314);
	outb(0x10, 0x313);
	outb(0x19, 0x310);
	outb(0x33, 0x311);
	outb(0x03, 0x313);
	outb(0xa3, 0x311);
}

/*
 * Initiation of btc cir keybaord
 */

void init_btc()
{
        outb(0x80, 0x313);
        outb(0x50, 0x314);
        outb(0x10, 0x313);
        outb(0x3d, 0x310);
        outb(0x5c, 0x311);
        outb(0x3d, 0x312);
        outb(0x5c, 0x314);
        outb(0x0c, 0x315);
        outb(0x31, 0x316);
        outb(0x03, 0x313);
        outb(0xa3, 0x311);
}

/*
 * Initiation of forward cir keyboard
 */
void init_forward()
{
        outb(0x80, 0x313);
        outb(0x00, 0x314);
        outb(0x10, 0x313);
        outb(0x31, 0x310);
        outb(0x46, 0x311);
        outb(0x11, 0x312);
        outb(0x26, 0x314);
        outb(0x17, 0x315);
        outb(0x23, 0x316);
        outb(0x17, 0x317);
        outb(0x24, 0x318);
        outb(0x33, 0x319);
        outb(0x41, 0x31a);
        outb(0x03, 0x313);
        outb(0xa3, 0x311);
}

static void cir_fixup(void)
{
        struct pci_dev *pcidev;

        pcidev = pci_find_device(0x1039, 0x0008, (void *)NULL);

        if (pcidev != NULL) {
            /* IO address for CIR */
            pci_write_config_byte(pcidev, 0x4A, 0x11);
            pci_write_config_byte(pcidev, 0x4B, 0x03);
            /* IRQ for CIR */
            pci_write_config_byte(pcidev, 0x6C, 0x05);
        }
}

static int __init bus_cir_init(void)
{
        pcibios_read_config_word(0,8,0x4a, &cir.base_reg);
        MSG("0:1.0 (0x4a) original base = %x\n", cir.base_reg);
        //pcibios_write_config_word(0,8,0x4a,0x311);

        pcibios_read_config_word(0,8,0x6c, &cir.irq_reg);
        MSG("0:1.0 (0x6c) original IRQ = %x\n", cir.irq_reg);
        //pcibios_write_config_word(0,8,0x6c,0x05);

	cir_fixup();
	if (check_region(CIR_BASE, CIR_EXTENT)) {
		cir.present = 0;
		return -EIO;
	}

	request_region(CIR_BASE, CIR_EXTENT, "cir");
	cir.present = 1;
	cir.active = 0;
	init_waitqueue_head(&cir.wait);
	cir.head = cir.tail = 0;
	data_no=0;

	MSG("SiS Comsumer IR detected, using IRQ %d.\n",cir_irq);

	misc_register(&bus_cir);

#ifdef SILITEK
        init_silitek();
#elif defined BTC
        init_btc();
#elif defined FORWARD
        init_forward();
#endif
        if (request_irq(cir_irq, cir_interrupt, SA_SHIRQ, "cir", &bus_cir)) {
                cir.active--;
                return -EBUSY;
        }
	MSG("CIR_ISR_PORT(0x312) : %x\n",inb(CIR_ISR_PORT));

	return 0;
}

void cir_clr(void)
{
        u16 ptr;

        //pcibios_read_config_dword(0,9,0,&ptr); //sis900

        pcibios_read_config_word(0,8,0x4A,&ptr);
        MSG("0:1.0 (0x4A) cir_fixup base =0x%x\n", ptr);

        pcibios_write_config_word(0,8,0x4A,cir.base_reg);
        pcibios_read_config_word(0,8,0x4A,&ptr);
        MSG("0:1.0 (0x4A) reset base to =0x%x\n", ptr);

        pcibios_read_config_word(0,8,0x6C,&ptr);
        MSG("0:1.0 (0x6C) cir_fixup IRQ =0x%x\n", ptr);

        pcibios_write_config_byte(0,8,0x6C,cir.irq_reg);
        pcibios_read_config_word(0,8,0x6C,&ptr);
        MSG("0:1.0 (0x6C) reset IRQ =0x%x\n", ptr);
}


#ifdef MODULE

int init_module(void)
{
	return bus_cir_init();
}

void cleanup_module(void)
{
	misc_deregister(&bus_cir);
        free_irq(cir_irq, &bus_cir);
        outb(0x0, 0x311);
        outb(0x3, 0x312);
	MSG("0x311 = %x,   0x312 = %x\n",inb(0x311), inb(0x312));
	cir_clr();
	release_region(CIR_BASE, CIR_EXTENT);
}
#endif
